#filter emptyLines

#define MOZ_LANGPACK_CREATOR The JavaScript Debugger Team

#define MOZ_LANGPACK_HOMEPAGE http://www.hacksrus.com/~ginda/venkman/

# If non-English locales wish to credit multiple contributors, uncomment this
# variable definition and use the format specified.
# #define MOZ_LANGPACK_CONTRIBUTORS <em:contributor>Joe Solon</em:contributor> <em:contributor>Suzy Solon</em:contributor>

#unfilter emptyLines
