/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Simon Fraser <smfr@smfr.org>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#import <AppKit/AppKit.h>

@class CertificateItem;
@class CertificateView;

@interface ViewCertificateDialogController : NSWindowController
{
  IBOutlet CertificateView*         mCertView;
  IBOutlet NSButton*                mOKButton;
  
  // runtime data
  BOOL                              mRunningModally;
  BOOL                              mAllowTrustSaving;    // whether to save the trust settings on window close
}


// if a window exists showing this cert, this will raise it and return its controller
+ (ViewCertificateDialogController*)showCertificateWindowWithCertificateItem:(CertificateItem*)inCertItem certTypeForTrustSettings:(unsigned int)inCertType;
// this will always run a new modal dialog
+ (void)runModalCertificateWindowWithCertificateItem:(CertificateItem*)inCertItem certTypeForTrustSettings:(unsigned int)inCertType;

// XXX we'll want to be able to custmize the buttons etc.

- (IBAction)defaultButtonHit:(id)sender;
- (IBAction)alternateButtonHit:(id)sender;

- (void)setCertificateItem:(CertificateItem*)inCert;
- (CertificateItem*)certificateItem;

- (void)setCertTypeForTrustSettings:(unsigned int)inCertType;  // one of the types in nsIX509Cert

- (int)runModally;
// default is YES
- (void)allowTrustSaving:(BOOL)inAllow;

@end

