//
//  SUPipedUnarchiver.h
//  Sparkle
//
//  Created by Andy Matuschak on 6/16/08.
//  Copyright 2008 Andy Matuschak. All rights reserved.
//

#ifndef SUPIPEDUNARCHIVER_H
#define SUPIPEDUNARCHIVER_H

#import <Cocoa/Cocoa.h>
#import "SUUnarchiver.h"

@interface SUPipedUnarchiver : SUUnarchiver {

}

@end

#endif
