//
//  SUUserInitiatedUpdateDriver.h
//  Sparkle
//
//  Created by Andy Matuschak on 5/30/08.
//  Copyright 2008 Andy Matuschak. All rights reserved.
//

#ifndef SUUSERINITIATEDUPDATEDRIVER_H
#define SUUSERINITIATEDUPDATEDRIVER_H

#import <Cocoa/Cocoa.h>
#import "SUUIBasedUpdateDriver.h"

@interface SUUserInitiatedUpdateDriver : SUUIBasedUpdateDriver {
	SUStatusController *checkingController;
	BOOL isCanceled;
}

@end

#endif
